import pandas as pd
from dataclasses import dataclass, field
from typing import TypedDict, List, Optional
from athina.interfaces.data import DataPoint
from pydantic import BaseModel


class EvalResultMetric(TypedDict):
    """
    Represents the LLM evaluation result metric.
    """

    id: str
    value: float


class DatapointFieldAnnotation(TypedDict):
    """
    The annotations to be logged for the datapoint field.
    """

    field_name: str
    text: str
    annotation_type: str
    annotation_note: str


class EvalResult(TypedDict):
    """
    Represents the LLM evaluation result.
    """

    name: str
    display_name: str
    data: dict
    failure: Optional[bool]
    reason: str
    runtime: int
    model: Optional[str]
    metrics: List[EvalResultMetric]
    datapoint_field_annotations: Optional[List[DatapointFieldAnnotation]]


@dataclass
class BatchRunResult:
    """
    Represents the result of a batch run of LLM evaluation.
    """

    eval_results: List[Optional[EvalResult]]
    eval_request_id: Optional[str] = field(default=None)

    def to_df(self):
        """
        Converts the batch run result to a Pandas DataFrame, including data and dynamic metrics.
        """
        pd.set_option("display.max_colwidth", 500)

        df_data = []
        for item in self.eval_results:
            if item is None:
                # Add a representation for None entries
                entry = {
                    "display_name": None,
                    "failed": None,
                    "grade_reason": None,
                    "runtime": None,
                    "model": None,
                    # Add more fields as None or with a placeholder as necessary
                }
            else:
                # Start with dynamic fields from the 'data' dictionary
                entry = {key: value for key, value in item["data"].items()}

                # Add fixed fields
                entry.update(
                    {
                        "display_name": item["display_name"],
                        "failed": item.get("failure"),
                        "grade_reason": item["reason"],
                        "runtime": item["runtime"],
                        "model": item.get("model"),
                    }
                )

                # Add dynamic metrics
                for metric in item["metrics"]:
                    entry[metric["id"]] = metric["value"]

            df_data.append(entry)

        df = pd.DataFrame(df_data)
        return df


class EvalPerformanceReport(TypedDict):
    """
    Represents the performance metrics for an evaluation.
    """

    true_positives: int
    false_positives: int
    true_negatives: int
    false_negatives: int
    accuracy: float
    precision: float
    recall: float
    f1_score: float
    runtime: int
    dataset_size: int


class GuardResult(BaseModel):
    passed: bool
    reason: str
    runtime: int



from typing import List, Optional

from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import answer_relevancy

"""
RAGAS Answer Relevancy Docs: https://docs.ragas.io/en/latest/concepts/metrics/answer_relevance.html
RAGAS Answer Relevancy Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_answer_relevance.py
"""
class RagasAnswerRelevancy(RagasEvaluator):
    """
    This evaluator focuses on assessing how pertinent the generated response is to the given prompt. 
    A lower score is assigned to responses that are incomplete or contain redundant information.
    """
    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_ANSWER_RELEVANCY.value

    @property
    def display_name(self):
        return "Ragas Answer Relevancy"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_ANSWER_RELEVANCY.value]
    
    @property
    def ragas_metric(self):
        return answer_relevancy
    
    @property
    def ragas_metric_name(self):
        return "answer_relevancy"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "context", "response"]

    @property
    def examples(self):
        return None
    
    @property
    def grade_reason(self) -> str:
        return "A response is deemed relevant when it directly and appropriately addresses the original query. Importantly, our assessment of answer relevance does not consider factuality but instead penalizes cases where the response lacks completeness or contains redundant details"

    def is_failure(self, score) -> Optional[bool]:
        return bool(score < self._failure_threshold) if self._failure_threshold is not None else None
        
    def generate_data_to_evaluate(self, query, context, response, **kwargs) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :param response: llm response
        :return: A dictionary with formatted data for evaluation
        """
        data = {
            "contexts": [context],
            "question": [query],
            "answer": [response]
        }
        return data



import os
from pydantic import BaseModel
from typing import List, Dict, Any, Optional
from athina.helpers.json import JsonExtractor
from athina.interfaces.model import Model
from athina.steps.base import Step
from athina.llms.abstract_llm_service import AbstractLlmService
from athina.keys import OpenAiApiKey
from athina.llms.openai_service import OpenAiService
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined
from athina.steps.transform import ExtractJsonFromString


class PromptMessage(BaseModel):
    role: str
    content: str

class ModelOptions(BaseModel):
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None

class ToolConfig(BaseModel):
    tool_choice: Optional[str] = None
    tools: Optional[List[Any]] = None

class PromptTemplate(BaseModel):
    messages: List[PromptMessage]
    env: Environment = None

    class Config:
        arbitrary_types_allowed = True
        
    @staticmethod
    def simple(message: str) -> "PromptTemplate":
        """Create a PromptTemplate from a string representation."""
        messages = [PromptMessage(role="user", content=message)]
        return PromptTemplate(messages=messages)

    def resolve(self, **kwargs) -> List[PromptMessage]:
        """Render the template with given variables."""

        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string='{{', 
            variable_end_string='}}',
            undefined=PreserveUndefined
        )
        resolved_messages = []
        for message in self.messages:
            content_template = self.env.from_string(message.content)
            content = content_template.render(**kwargs)
            resolved_message = PromptMessage(role=message.role, content=content)
            resolved_messages.append(resolved_message)

        return resolved_messages


class PromptExecution(Step):
    """
    Step that executes a prompt using an LLM service.

    Attributes:
        llm_service (AbstractLlmService): The LLM service to use for prompt execution.
        template (PromptTemplate): The template to render the prompt.
        model (str): The model to use for the LLM service.
    """

    llm_service: AbstractLlmService = None
    template: PromptTemplate
    model: str
    model_options: ModelOptions
    tool_config: Optional[ToolConfig] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if kwargs.get("llm_service"):
            self.llm_service = kwargs.get("llm_service")
        else:
            self.llm_service = OpenAiService()

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    def simple(message: str, model: str = Model.GPT4_O.value) -> "PromptExecution":
        OpenAiApiKey.set_key(os.getenv("OPENAI_API_KEY"))
        openai_service = OpenAiService()
        return PromptExecution(
            llm_service=openai_service,
            template=PromptTemplate.simple(message),
            model=model,
        )

    def execute(self, input_data: dict, **kwargs) -> str:
        """Execute a prompt with the LLM service."""
        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict) and self.input_key:
            raise ValueError("PromptExecution Error: Input data must be a dictionary")

        try:
            response = self.template.resolve(**input_data)
            response = self.llm_service.chat_completion(
                response, model=self.model, **self.model_options.model_dump(), **(self.tool_config.model_dump() if self.tool_config else {})
            )
            output_type = kwargs.get('output_type', None)
            error = None
            if output_type:
                if output_type == "string":
                    if not isinstance(response, str):
                        error = "LLM response is not a string"
                elif output_type == "array":
                    extracted_response = ExtractJsonFromString().execute(response)
                    if not isinstance(extracted_response, list):
                        error = "LLM response is not an array"
                    response = extracted_response

                elif output_type == "object":
                    extracted_response = ExtractJsonFromString().execute(response)
                    if not isinstance(extracted_response, dict):
                        error = "LLM response is not an object"
                    response = extracted_response

            elif not isinstance(response, str):
                error = "LLM service response is not a string"

            if error:
                return {
                    "status": "error",
                    "data": error
                }
            else: 
                return {
                    "status": "success",
                    "data": response
                }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e)
            }


from typing import Union, Dict, Iterable, Any
from athina.helpers.json import JsonExtractor
from athina.steps import Step


class ExtractJsonFromString(Step):
    """
    Step that extracts JSON data from a string.
    """

    def execute(
        self, input_data: str
    ) -> Union[Dict[str, Any], Iterable[Dict[str, Any]]]:
        """Extract JSON data from the input string."""

        if (
            input_data is None
            or not isinstance(input_data, str)
            or len(input_data) == 0
        ):
            raise TypeError("Input data must be a valid string.")

        output = JsonExtractor.extract_first_json_entity(input_data)

        if output is None:
            raise ValueError("No valid JSON data found in the input string.")
        return output



from typing import Optional
from athina.loaders import ResponseLoader, Loader


class LoaderHelper:
    """Helper class for loading data"""

    @staticmethod
    def get_loader(eval_name, loader_name: Optional[str] = None):
        """Returns the loader for the given format"""
        if eval_name == "ContextContainsEnoughInformation" or eval_name == "DoesResponseAnswerQuery" or eval_name == "Faithfulness":
            return Loader
        else:
            if loader_name is None:
                raise ValueError(
                    f"Loader name must be specified for {eval_name} evaluation."
                )
            else:
                return ResponseLoader

    @staticmethod
    def load(eval_name, format, **kwargs):
        """Loads data based on the format specified."""
        loader = LoaderHelper.get_loader(eval_name)
        return loader().load(format, **kwargs)



class AthinaMessages:
    """
    User facing messages.
    """

    SIGN_UP_FOR_BEST_EXPERIENCE = """
For the best experience, sign up at https://athina.ai and set an Athina API key.

See https://docs.athina.ai/evals/quick_start for more information.
"""
    
    NO_ATHINA_API_KEY = """
Please set an Athina API key.

See https://docs.athina.ai/evals/quick_start for more info.
    """

    NO_OPENAI_API_KEY = """
Please set an OpenAI API key.

See https://docs.athina.ai/evals/quick_start for more info.
    """




import math
from abc import abstractmethod
from typing import Optional
from athina.interfaces.athina import AthinaExperiment
from athina.interfaces.model import Model
import time
from typing import Optional, Any, List
from athina.interfaces.result import EvalResult, EvalResultMetric
from athina.interfaces.model import Model
from athina.helpers.logger import logger
from ..base_evaluator import BaseEvaluator
from datasets import Dataset
from langchain_openai.chat_models import ChatOpenAI
from ragas.llms import LangchainLLM
from ragas import evaluate
from athina.keys import OpenAiApiKey


class RagasEvaluator(BaseEvaluator):
    _model: str
    _openai_api_key: Optional[str]
    _experiment: Optional[AthinaExperiment] = None
    _failure_threshold: Optional[float] = None

    def __init__(
        self,
        openai_api_key: Optional[str] = None,
        model: Optional[str] = None,
        failure_threshold: Optional[float] = None
    ):
        if model is None:
            self._model = self.default_model
        else:
            self._model = model
        
        if openai_api_key is None:
            self._openai_api_key = OpenAiApiKey.get_key()
        else:
            self._openai_api_key = openai_api_key

        if failure_threshold is not None:
            self._failure_threshold = failure_threshold

    @property
    def default_model(self) -> str:
        return Model.GPT35_TURBO.value
    
    def generate_data_to_evaluate(self, **kwargs):
        pass
    
    @abstractmethod
    def ragas_metric(self) -> Any:
        pass

    @property
    def grade_reason(self) -> str:
        raise NotImplementedError
    
    def _get_model(self):
        return ChatOpenAI(model_name=self._model, api_key=self._openai_api_key)

    def _evaluate(self, **kwargs) -> EvalResult:
        """
        Run the Ragas evaluator.
        """
        start_time = time.time()
        self.validate_args(**kwargs)
        metrics = []
        try:
            self.ragas_metric.llm = LangchainLLM(llm=self._get_model())
            data = self.generate_data_to_evaluate(**kwargs)
            dataset = Dataset.from_dict(data)
            scores = evaluate(dataset, metrics=[self.ragas_metric])
            metric_value = scores[self.ragas_metric_name]
            if isinstance(metric_value, (int, float)) and not math.isnan(metric_value):
                metrics.append(EvalResultMetric(id=self.metric_ids[0], value=metric_value))
            else:
                logger.warn(f"Invalid metric value: {metric_value}")

            failure = self.is_failure(score=metric_value)
        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        llm_eval_result = EvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            failure=failure,
            reason=self.grade_reason,
            runtime=eval_runtime_ms,
            model=self._model,
            metrics=metrics,
        )
        return {k: v for k, v in llm_eval_result.items() if v is not None}





SYSTEM_MESSAGE = """You are given a list of messages from a conversation, with each message in the order it was sent. 

Your task is to analyze the flow of messages by the AI. For every message by the AI, follow these steps:

1. Read the message and consider it in the context of the previous messages in the conversation.

2. Think about the following:
- Does this message logically follow from the previous ones?
- Is there any contradiction or sudden shift in topic that makes this message seem out of place?

3. Decide if the message is logically "coherent" (it logically follows the conversation so far) or "not_coherent" (it breaks the logical flow or contradicts previous messages).

After considering each AI message through these steps, record your evaluation in a JSON object like this:

{ 
    "details": [ 
        {
            "message": message1,
            "result": "coherent / not_coherent",
            "explanation": “explanation of why this message is or is not coherent w.r.t previous messages"
        },
        ...
    ]
}

You must evaluate every single message in the conversation.
"""

USER_MESSAGE = """
Here is the conversation you need to evaluate:
{messages}
"""
